#!/usr/bin/python3

#	cve-manager : CVE management tool
#	Copyright (C) 2017-2022 Alexey Appolonov
#
#	This program is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <http://www.gnu.org/licenses/>.

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

from collections        import defaultdict
from cve_manager.common import Init
from cve_manager.conf   import DB_CON_SEC
from cve_manager.db     import DB
from cve_manager.url    import ParseURL, JoinURL

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Interface for the vulnerability database

class Query():

	CHART_SIZE = 50

	PRODUCTS_COND = "WHERE vendor <> '' AND product <> '' " \
		"ORDER BY vendor, product"

	def __init__(self, args):

		self.tables = tuple(f'{branch}_src' for branch in args.branches)
		self.printer, conf = Init(args, monitor=True)
		self.mysql_config, = conf.Get([DB_CON_SEC])

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Send given query to the DB

	def __Q(self, q, tables=tuple(), printout=True):

		db = DB(self.printer)
		if not db.Connect(self.mysql_config):
			exit(1)

		cursor = db.Cursor(q, tables=tables)
		if cursor == None:
			exit(1)

		res = cursor.fetchall()

		if printout:
			for field in res:
				print(', '.join(field))

		return res

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Query package names of all branches

	def Packages(self, additional_fileds=[], condition='', printout=True):

		fields = ', '.join(['name'] + additional_fileds)
		union = f' UNION SELECT {fields} FROM '.join(self.tables)
		q = f"SELECT * FROM (SELECT DISTINCT {fields} FROM {union}) alias " \
			f"WHERE name <> '' {f'AND ({condition})' if condition else ''} " \
			"ORDER BY name"

		return self.__Q(q, tales=self.tables, printout=printout)

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Query vendor names and product names used by NVD

	def CPEs(self, printout=True):

		tables = ('cpe_import', 'nvd_products')
		union = " UNION SELECT vendor, product FROM ".join(tables)
		q = "SELECT * FROM " \
				f"(SELECT DISTINCT vendor, product FROM {union}) alias " + \
				+ self.PRODUCTS_COND

		return self.__Q(q, tables=tables, printout=printout)

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Query vendor names and product names used by FSTEC

	def FSTECProducts(self, printout=True):

		tables = ('fstec_products',)
		q = f"SELECT vendor, product FROM {tables[0]} " + self.PRODUCTS_COND

		return self.__Q(q, tables=tables, printout=printout)

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Query package names and product names of mapped packages of all branches
	# and return a dict {package_name: [product_name_from_branch_i, ..
	# .. , product_name_from_branch_j]}

	def ReverseIntegralBranchMap(self):

		reverse_branch_map = defaultdict(list)
		union = " UNION SELECT name, nvd_name FROM ".join(self.tables)
		q = f"SELECT * FROM (SELECT DISTINCT name, nvd_name FROM {union}) alias " \
			"WHERE name <> '' AND nvd_name <> ''"

		for name, nvd_name in self.__Q(q, tables=self.tables, printout=False):
			reverse_branch_map[nvd_name].append(name)

		return reverse_branch_map

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Query vendor names and product names and return a dict
	# {vendor: {product_names}}

	def ProductsOfVendors(self):

		products_of_vendors = defaultdict(set)

		for vendor, product in self.CPEs(printout=False):
			products_of_vendors[vendor].add(product)

		return products_of_vendors

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Query all refs from nvd_vul_import table, count the number of occurances
	# of each network location and each url path (network location + path) and
	# print the charts

	def __PrintChart(self, data):

		print('=' * 80)

		total_count = sum(data.values())

		if total_count > 0:
			chart = sorted(data.items(), key=lambda kv: kv[1], reverse=True) \
				[:self.CHART_SIZE]
			for item, count in chart:
				print(f'{item}: {count} ({count * 100 / total_count:.2f}%)')

		print()


	def NVDVulImportRefs(self):

		netlocs = defaultdict(int)
		urls = defaultdict(int)
		tables = ('nvd_vul_import',)
		q = f"SELECT refs from {tables[0]} " \
			"WHERE refs <> '' AND refs IS NOT NULL"

		for refs, in self.__Q(q, tables=tables, printout=False):
			netlocs_of_current_refs = set()
			parsed_urls_of_current_refs = set()
			for ref in refs.split():
				url = ParseURL(ref)
				if not url['netloc']:
					continue
				netlocs_of_current_refs.add(url['netloc'])
				parsed_urls_of_current_refs.add(JoinURL([url['netloc'], url['path']]))
			for netloc in netlocs_of_current_refs:
				netlocs[netloc] += 1
			for parsed_url in parsed_urls_of_current_refs:
				urls[parsed_url] += 1

		self.__PrintChart(netlocs)
		self.__PrintChart(urls)

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Query all product names

	def ProductNames(self):

		return set(product for _, product in self.CPEs(printout=False))
